#ifndef RMDP_RMDP_H
#define RMDP_RMDP_H
#include "compat.h"

#ifdef DEBUGE
#define DEB(x) {x;}
#else
#define DEB(x)
#endif
#define NDEB(x)

#define MAX_RMDP_INSTANCES	128

#define RMDP_PKT_SIZE	256
#define RMDP_BASE_RATE	16*1024
#define RMDP_NLAY	5

#define THE_K		8

#define RMDP_MAX_RETRIES	5	/* maximum # of continuation requests
					 * before aborting */
#define RMDP_LAST_RETR_TIM	1000000

/*++++
 * Data packet (actually is RLC payload)
 */
 
typedef struct _rmdp_header_t {
	unsigned short B;
	unsigned short k;
	unsigned short the_B;
	unsigned short the_k;
	unsigned long sessid;
	unsigned long flength;		/* in bytes !! */
	unsigned long left;		/* left in the base layer */
} rmdp_header_t;

#define RMDP_H_S	sizeof(rmdp_header_t)
#define RMDP_DATA(p)	((void *)((char *)(p)+RMDP_H_S))
#define RMDP_DATA_2_P(p)	((void *)((char *)(p)-RMDP_H_S))
#define RMDP_H_P(p)	((rmdp_header_t *)(p))

/*++++
 * Continuation request packet (UDP payload)
 * and other feedback...
 */
 
#define COMM_FEEDHEAD							       \
	unsigned long magic;						       \
	unsigned long sessid;						       \
	unsigned long extension;	/* packet on the base layer */	       \
	unsigned long wasleft;		/* left at the req. time */

typedef struct _rmdp_cont_pkt_t {
	COMM_FEEDHEAD
} rmdp_cont_pkt_t;

typedef struct _rmdp_feed_pkt_t {
	COMM_FEEDHEAD
	unsigned long type;
} rmdp_feed_pkt_t;

#define RMDP_STATS_PAYSIZE 1024

#define RMDP_L_MAGIC	0x15a85afe
#define RMDP_FEED_RLC_STATS	0x15af33dc

#define RMDP_CP_S	sizeof(rmdp_cont_pkt_t)
#define RMDP_CP_P(p)	((rmdp_cont_pkt_t *)(p))
#define RMDP_CP_FILL(p, sid, ext, lef)			\
	{						\
		(p).magic =	htonl(RMDP_L_MAGIC);	\
		(p).sessid =	htonl(sid);		\
		(p).extension =	htonl(ext);		\
		(p).wasleft =	htonl(lef);		\
	}

#define RMDP_FEE_RLC_FILL(p, sid)			\
	{						\
		RMDP_CP_FILL(p, sid, 0, 0);		\
		(p).type = htonl(RMDP_FEED_RLC_STATS);	\
	}

#endif /* RMDP_RMDP_H */
