#ifndef RMDP_RMDPIF_H
#define RMDP_RMDPIF_H

#define RMDP_LEFT_INFINITE 0xffffffff

#define RMDP_BASE_ERROR	-200
#define RMDP_ERROR_SIZE 17
/*
** Error codes.
*/
#define RMDP_INITERR		-200
#define RMDP_STAT_ERROR		-201
#define RMDP_NOMCAST_ADDR	-202
#define RMDP_COMPUTE_K		-203
#define RMDP_TOOMANY_INST	-204
#define RMDP_NOSUCH_INST	-205
#define RMDP_ALLOC_ERROR	-206
#define RMDP_DATA_SIZE		-207
#define RMDP_NO_UI		-208
#define RMDP_ER_RETR_TI		-209
#define RMDP_ER_TIMEOUT		-210
#define RMDP_ER_SND		-211
#define RMDP_NO_SOCK		-212
#define RMDP_OUT_FILE_ERR	-213
#define RMDP_COD_INST_ERR	-214
#define RMDP_DECODER_ERR	-215
#define RMDP_NO_RLC_ID		-216


#define RMDP_BASE_WARNING 2000
#define RMDP_WARNING_SIZE 5
/*
** Warning codes.
*/
#define RMDP_NOERR_H		2000	/* Warning: no error messages */
#define RMDP_CANT_READ_REQ	2001	/* Warning: cannot read request */
#define RMDP_BAD_REQ		2002	/* Warning: bad request messages */
#define RMDP_TIMEOUT_WAR	2003	/* reception timeout warning */
#define RMDP_CANT_WRITE_STATS	2004	/* cannot write stats warning */

typedef struct _rmdp_addr_t {
    unsigned long addr;
    short port;
    short ttl;
} rmdp_addr_t;

typedef int (*rmdp_tx_packet_t)
	(int pi, int lay, char flags, char **p);

typedef void (*rmdp_rx_packet_t)
	(int pi, int lay, char flags, char *p, int size);

/** called when when the reception/transmission terminates successfully **/
typedef void (*rmdp_end_f)
	(int pi, int result);

/** called when when the reception terminates because of a timeout **/
typedef void (*rmdp_rx_err_f)
	(int pi);

/*
** Main routines ...
*/
int rmdp_new (int isreceiver);
int rmdp_set_addr (int pi, rmdp_addr_t *adds);
int rmdp_set_io (int pi, int fd);
int rmdp_set_id (int pi, unsigned long id);
int rmdp_get_id (int pi, unsigned long *id);
int rmdp_get_f (int pi, void **f);
int rmdp_delete (int pi);
int rmdp_get_psize (int pi);
int rmdp_get_rlc (int pi, unsigned long *id);
int rmdp_set_rlc (int pi, int rlc_pi);
int rmdp_enable_gui (int pi);
int rmdp_set_name (int pi, char *name);
int rmdp_set_end_callback (int pi, rmdp_end_f f);
int rmdp_set_stat_reports (int pi, int stafd);

/*
** sender side routines
*/
int rmdp_get_nlay (int pi);
int rmdp_get_left (int pi);
int rmdp_incr_left (int pi, int incr);

/*
** receiver side routines
*/
int rmdp_rx_set_err_callback (int pi, rmdp_rx_err_f f);
int rmdp_rx_enable_multifeed (int pi, int multifeed);
/* ... must add gui stuff */

/*
** utils function
*/
int rmdp_parse_addr(char *s, char *a, int addrlen, short *p, char *msg);
int rmdp_n_parse_addr(char *s, char *a, int addrlen, short *p, int *ttl,
	char *msg);
#if __STDC__
void rmdp_my_error (int rmdp_i, char const *fmt, ...);
#else
void rmdp_my_error (rmdp_i, fmt, va_alist)
  int rmdp_i;
  char *fmt;
  va_dcl;
#endif


#endif /* RMDP_RMDPIF_H */
