#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>
#include <fcntl.h>

#include <rlcif.h>

#include "rmdpif.h"
#include "fec.h"
#include "rmdp.h"
#include "layers.h"
#include "sender.h"
#include "agent.h"
#include "init.h"
#include "utils.h"

#if (!defined(Linux))
#if ( defined(SunOS) && (OSMVER < 5))
extern char *sys_errlist[];
extern int errno;
#endif
#endif

/* this is an hack ... I'm using rlc net.o */
int openrsock(unsigned long addr, short *port);
unsigned long localaddr(int ssock_);

static int get_flen(int filn);
static int setup_req_chan(rmdp_inst_t *se);
#define getinstance(i) ((rmdp_inst_t *)_rlc_getinstance(i, rmdp_inst_mngr))

#define ID_RMDP_REQU 212134

static int sender_init_done = 0;

/*++++
 Retrieves the number of layers used in the sender.

Params:
    - pi: protocol instance identifier.
 
Returns:
    - the number of layers, on success.
    - a negative value on failure:
*/
int
rmdp_get_nlay (int pi)
{
  rmdp_inst_t *inst;
 
    if ((inst = getinstance(pi)) == NULL) {
        return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }

    return ( inst->nlay );
}

/*++++
 Retrieves the number of packets left to transmit in the base layer.

Params:
    - pi: protocol instance identifier.
 
Returns:
    - the number of packets, on success.
    - a negative value on failure:
*/
int
rmdp_get_left (int pi)
{
  rmdp_inst_t *inst;
 
    if ((inst = getinstance(pi)) == NULL) {
        return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }

    return ( inst->pkts_left );
}

/*++++
 Increments the number of packets left to transmit in the base layer.

Params:
    - pi: protocol instance identifier.
 
Returns:
    - 0 on success.
    - a negative value on failure:
*/
int
rmdp_incr_left (int pi, int incr)
{
  rmdp_inst_t *inst;
 
    if ((inst = getinstance(pi)) == NULL) {
        return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }

    inst->pkts_left = inst->pkts_left + incr;

    return ( 0 );
}

int _rmdp_sender_init()
{
    if (!sender_init_done) {
	sender_init_done = 1;
    }

    return (0);
}

/*
 * returns the data packet to send and its lenght,
 * only the payload is filled, the transport header gets filled
 * by the caller.
 * 
 */
int
_rmdp_tx_packet(int des, int lay, char flags, char **p)
{
  int pkt, blk, res;
  rmdp_inst_t *inst;
  rmdp_header_t *mp;

    if ((inst = getinstance(des)) == NULL) {
	rlc_have_error_(-1, RMDP_NOSUCH_INST, NULL);
	return (RMDP_NOSUCH_INST);
    }
	
    if (inst->mfeeso == -1 && (res = setup_req_chan(inst)) < 0) {
	_rmdp_deleteinstance(inst);
	rlc_have_error_(inst->rlc_pi, res, NULL);
    }

    /* lay < 0 is a termination */
    if (lay >= 0 && inst->pkts_left>0) {

	which_packet(inst->B, inst->k, inst->nlay, -1,
	    &lay, &(inst->idxs[lay]), &pkt, &blk);
	(inst->idxs[lay]) ++;
	if (pkt < inst->k)
	    *p = (char *)mp = (rmdp_header_t *)inst->alldata[blk][pkt];
	else {
	    fec_encode(inst->feccode, (void **)inst->allpkts[blk],
		RMDP_DATA(inst->sparepkt), pkt, inst->psize);
	    *p = (char *)mp = (rmdp_header_t *)inst->sparepkt;
	}
	mp->B = htons(inst->B);
	mp->k = htons(inst->k);
	mp->the_B = htons(blk);
	mp->the_k = htons(pkt);
	mp->sessid = htonl(inst->sessid);
	mp->flength = htonl(inst->flength);
	mp->left = htonl(inst->pkts_left);

	if (!lay && inst->pkts_left != RMDP_LEFT_INFINITE) {
	    inst->pkts_left --;
	}
	NDEB(fprintf(stderr, "&&&& left %ld\n", inst->pkts_left);)
	return(inst->psize+RMDP_H_S);
    } else {
	_rmdp_deleteinstance(inst);
	return(0);
    }
}

/*
 * Reads the file to be sent.
 * 
 */
int
_rmdp_tx_read(int des, int fd)
{
  rmdp_inst_t *inst;
  int tlength, done, i, j;

    if ((inst = getinstance(des)) == NULL) {
        return (rlc_return_error_(-1, RMDP_NOSUCH_INST, NULL));
    }
	
    /* XXX we should make sure that this get called just once
     * or at least .... */
    if ((tlength = get_flen(fd)) <= 0)
	return rlc_return_error_(inst->rlc_pi, RMDP_STAT_ERROR, NULL);
    inst->flength = tlength;
    inst->npkts = (tlength+inst->psize-1)/inst->psize;

    if (compute_k_B(&(inst->k), &(inst->B), inst->npkts, inst->nlay) < 0) {
	return (rlc_return_error_ (inst->rlc_pi, RMDP_COMPUTE_K, NULL));
    }
    if ((inst->feccode = get_fec(inst->k, 256)) == NULL) {
	return (rlc_return_error_ (inst->rlc_pi, RMDP_COD_INST_ERR, NULL));
    }
    /*
    ** computing the number of packets scheduled initially
    ** tricky... we use the maximum between those sufficient
    ** at an average layer and 40 ...
    */
    inst->init_left = max (40, inst->k*inst->B / (1<<((inst->nlay)/2)) );
#if 0
    inst->init_left *= 1000;	/* XXX */
    NDEB(fprintf(stderr, "&&&& left %ld\n", inst->init_left);)
#endif
    if ( inst->pkts_left != RMDP_LEFT_INFINITE)
	inst->pkts_left = inst->init_left;

    /* allocate file buffers and read file */
    if ( _rmdp_allocation(inst) < 0 )
	return (rlc_return_error_(inst->rlc_pi, RMDP_ALLOC_ERROR,
	    "* rmdp: Cannot allocate file data structure"));
    for (done=0, i=0; i<(inst->B); i++) {
	for (j=0; j<(inst->k); j++) {
	    if (NULL ==(inst->alldata[i][j]  =
		(char *)calloc(1, inst->psize+RMDP_H_S))) {
		return (rlc_return_error_(inst->rlc_pi, RMDP_ALLOC_ERROR,
		    "* rmdp: Cannot allocate file data structure"));
	    }
	    inst->allpkts[i][j] = RMDP_DATA(inst->alldata[i][j]);
	    if (!done && (read(fd, inst->allpkts[i][j], inst->psize)) !=
		inst->psize) {
		done = 1;
	    }
	}
    }

    return 0;
}

static void
have_extension_req(rmdp_inst_t *se, rmdp_cont_pkt_t *req)
{
  unsigned long ext;

    ext = ntohl(req->extension);
    NDEB(fprintf(stderr, "> req %ld, now %ld!\n", ext, se->pkts_left);)
    if (ext > 0  && se->pkts_left != RMDP_LEFT_INFINITE)
	/* XXX we should implement more sophisticated policies */
	se->pkts_left += min (ext, se->init_left);
}

static int
have_feedback(rmdp_inst_t *se)
{
  char feedpkt[4096];
  rmdp_cont_pkt_t *extreq;
  rmdp_feed_pkt_t *feereq;
  struct sockaddr_in srcaddr;
  int len, srcaddrlen=sizeof(srcaddr);
  int sock;

    sock = rlc_showselected ();
    
    NDEB(fprintf(stderr, "> have feedback!\n");)
    /* read request */
    if ((len = recvfrom(sock, (char *)&feedpkt, sizeof(feedpkt), 0,
	(struct sockaddr *)&srcaddr, &srcaddrlen))<=0 ) {
	/* ... this is a warning */
	rlc_have_error_ (se->rlc_pi, RMDP_CANT_READ_REQ, NULL);
	NDEB(fprintf(stderr, "> have request cant read *** !\n");)
	return 0;
    }

    extreq = (rmdp_cont_pkt_t *)&feedpkt;
    /* check integrity */
    if (len < sizeof(rmdp_cont_pkt_t) ||
	ntohl(extreq->magic) != RMDP_L_MAGIC ||
	ntohl(extreq->sessid) != se->sessid) {
	/* ... this is a warning */
	rlc_have_error_ (se->rlc_pi, RMDP_BAD_REQ, NULL);
	NDEB(fprintf(stderr, "> have request bad req *** !\n");)
	return 0;
    }

    if (ntohl(extreq->extension) != 0)
	have_extension_req (se, extreq);

    if (len >= sizeof(rmdp_cont_pkt_t) + 4) {
					/* there's something else */
	feereq = (rmdp_feed_pkt_t *)&feedpkt;
	switch (ntohl(feereq->type)) {
	case RMDP_FEED_RLC_STATS:
	    if (se->stafd > 0) {
		stats_rec_t st;
		int size = len - sizeof(rmdp_cont_pkt_t) - 4;
		st.time = htonl(timestamp_s());
		st.fromip = srcaddr.sin_addr.s_addr;
		st.fromport = srcaddr.sin_port;
		st.type = feereq->type;
		st.size = htonl(size);
		if (write(se->stafd, (void *)&st, sizeof(st)) != sizeof(st) ||
		    write(se->stafd, feedpkt + len - size, size) != size) {
		    se->stafd = -1;
		    rlc_have_error_ (se->rlc_pi, RMDP_CANT_WRITE_STATS, NULL);
		}
	    }
	    break;
	default:			/* don't know this, ignore */
	    break;
	}
    }

    return 0;
}

static int setup_req_chan(rmdp_inst_t *se)
{
  unsigned long myaddr;
    /* first arrange for multicast feedback */
    if ((se->mfeeso = openrsock(se->addr, &(se->port))) < 0 ||
	fcntl(se->mfeeso, F_SETFL, O_NONBLOCK) < 0 ||
	rlc_insertchan(se->mfeeso, ID_RMDP_REQU, (CALLBACK)have_feedback,
	(void *)se) < 0) {
	return (rlc_internal_set_error_ (se->rlc_pi, RMDP_NO_SOCK, NULL));
    }

    /* then for unicast ... */
    if ((myaddr = localaddr(se->mfeeso)) == -1)
	return (rlc_internal_set_error_ (se->rlc_pi, RMDP_NO_SOCK, "Cannot"
		"determine local address"));

    if ((se->ufeeso = openrsock(myaddr, &(se->port))) < 0 ||
	fcntl(se->ufeeso, F_SETFL, O_NONBLOCK) < 0 ||
	rlc_insertchan(se->ufeeso, ID_RMDP_REQU, (CALLBACK)have_feedback,
	(void *)se) < 0) {
	return (rlc_internal_set_error_ (se->rlc_pi, RMDP_NO_SOCK, NULL));
    }

    return 0;
}

static int get_flen(int filn)
{
  struct stat filest;
  int i=fstat(filn, &filest);
    if (i < 0) {
#if (!defined(Linux))
#if ( defined(SunOS) && (OSMVER < 5))
	return rlc_return_error_(-1, RMDP_STAT_ERROR, sys_errlist [errno]);
#else
	return rlc_return_error_(-1, RMDP_STAT_ERROR, strerror (errno));
#endif
#else
	return rlc_return_error_(-1, RMDP_STAT_ERROR, strerror (errno));
#endif
    }
    return(filest.st_size);
}

